###################################################################################################
###---------------------------------------------------------------------------------------------###
### This file is part of CPB (Correlated Pattern Biclustering Algorithm)                        ###
### Copyright (c) 2011,                                                                         ###
### By:    Kemal Eren,                                                                          ###
###        Mehmet Deveci,                                                                       ###
###        Onur Kucuktunc,                                                                      ###
###        Kemal Eren,                                                                          ###
###        Doruk Bozdag,                                                                        ###
###        Umit V. Catalyurek                                                                   ###
###                                                                                             ###
###---------------------------------------------------------------------------------------------###
### For license info, please see the README.txt and LICENSE.txt files in the main directory.    ###
###---------------------------------------------------------------------------------------------###
###################################################################################################

"""
Classes for represting biclusters, and some utility functions for dealing with
common bicluster tasks, like IO.

"""

class Bicluster:
    """A class for representing biclusters."""


    def __init__(self, rows, cols, data=None):
        """
        @param int[] rows: the row indices that make up this bicluster.
        @param int[] cols: the column indices that make up this bicluster.
        @param float[][] data: full dataset to which this bicluster refers. Necessary for certain methods. It is in 2 dimensional numpy.array format.

        """
        self.rows = rows
        self.cols = cols
        self.data = data


    def __eq__(self, other):
        """
        Test two biclusters for equality.
        
        Two biclusters are equal if the have the same rows and columns, and they have
        the same object as their data member. It is not enough that their data be
        equal; it must be the same object.
                
        @param Bicluster other: the given bicluster to compare.

        """

        return set(self.rows) == set(other.rows) and \
               set(self.cols) == set(other.cols) and \
               id(self.data) == id(other.data)


    def get_array(self, rows=None, cols=None):
        """
        Get a numpy array bicluster from data, using the indices in bic_indices.

        Note: requires that this Bicluster's data member is not None.
        
        @param int[] rows: the row indices.
        @param int[] cols: the column indices .

        """
        if not self.data is None:
            if rows is None:
                rows = self.rows
            if cols is None:
                cols = self.cols
            array = self.data.take(rows, axis=0).take(list(cols), axis=1)
            return array


    def intersection(self, other):
        """
        Returns a new bicluster with common rows and columns.

        Note: Returned bicluster does not have its data member set.
        @param Bicluster other: the given bicluster to do intersection operation.
        """

        rows = set(self.rows).intersection(set(other.rows))
        cols = set(self.cols).intersection(set(other.cols))
        return Bicluster(rows, cols)


    def union(self, other):
        """
        Returns a new bicluster with union of rows and columns.

        Note: Returned bicluster does not have its data member set.
        @param Bicluster other: the given bicluster to do union operation.

        """

        rows = set(self.rows).union(set(other.rows))
        cols = set(self.cols).union(set(other.cols))
        return Bicluster(rows, cols)


    def symmetric_difference(self, other):
        """
        Returns a new bicluster with only unique rows and columns.

        i.e. if self.rows == [3, 4] and other.rows = [4, 5], then the
        new bicluster will have rows == [3, 5].

        Note: Returned bicluster does not have its data member set.
        @param Bicluster other: the given bicluster to do difference operation. 
        """

        rows = set(self.rows).symmetric_difference(set(other.rows))
        cols = set(self.cols).symmetric_difference(set(other.cols))
        return Bicluster(rows, cols)
   

    def area(self):
        """Returns the number of data points in this bicluster."""
        return len(self.rows) * len(self.cols)

    def overlap(self, other):
        """
        Returns the ratio of the overlap size with the given bicluster (other) to the total self size.
        @param Bicluster other: the bicluster to check overlap with.
        """
        return self.intersection(other).area() / self.area()

    def difference(self,other):
        """
        Returns the difference of two biclusters.
        The returned bicluster has the rows and the columns that does not belong to the given bicluster (other).
        @param Bicluster other: the given bicluster to do difference operation.
        """
        rows = set(self.rows).difference(set(other.rows))
        cols = set(self.cols).difference(set(other.cols))
        return Bicluster(rows, cols)

    def __repr__(self):
        return "Bicluster({0}, {1})".format(repr(self.rows), repr(self.cols))

    def issubset(self, other):
        return set(self.rows).issubset(set(other.rows)) and set(self.cols).issubset(set(other.cols))



def filter_biclusters(biclusters, min_rows=2, min_cols=2, max_overlap=1.0, remove_subsets=True):
    """
    Removes duplicates, small biclusters, and overlapping biclusters from a list.
    
    :param min_rows: the minimum allowed rows
    :param min_cols: the minimum allowed columns
    :param max_overlap: the maximum allowed % overlap between any two clusters; between 0 and 1

    """
    assert max_overlap >=0 and max_overlap <= 1
    if min_rows >= 0 and min_cols >= 0:
        #remove empty biclusters and biclusters as big as the dataset
        biclusters = [b for b in biclusters if len(b.rows) > min_rows and len(b.cols) > min_cols
                  and (len(b.rows) < b.data.shape[0] or len(b.cols) < b.data.shape[1])]

    biclusters = sorted(biclusters, cmp = lambda x, y: cmp(x.area(), y.area()), reverse=True)

    def keep(b, accepted):
        for a in accepted:
            if b == a or b.overlap(a) > max_overlap or (remove_subsets and b.issubset(a)):
                return False
        return True

    accepted = []
    for b in biclusters:
        if keep(b, accepted):
            accepted.append(b)
    return accepted


def write_biclusters(biclusters, filename):
    """
    Write biclusters to an output file.
    
    Uses the format:

    <rows>                                          \n
    <cols>                                          \n

    seperated by empty lines
    @param Bicluster[] biclusters: the list of biclusters that will be written to the file
    @param string filename: the name of the output file.
    """

    with open(filename, 'w') as outfile:
        for bicluster in biclusters:
            outfile.write(' '.join([str(r) for r in bicluster.rows]))
            outfile.write("\n")
            outfile.write(' '.join([str(c) for c in bicluster.cols]))
            outfile.write("\n")
            outfile.write("\n")
    outfile.close()
