###################################################################################################
###---------------------------------------------------------------------------------------------###
### This file is part of CPB (Correlated Pattern Biclustering Algorithm)                        ###
### Copyright (c) 2011,                                                                         ###
### By:    Kemal Eren,                                                                          ###
##         Mehmet Deveci,                                                                       ###
###        Onur Kucuktunc,                                                                      ###
###        Doruk Bozdag,                                                                        ###
###        Umit V. Catalyurek                                                                   ###
###                                                                                             ###
###---------------------------------------------------------------------------------------------###
### For license info, please see the README.txt and LICENSE.txt files in the main directory.    ###
###---------------------------------------------------------------------------------------------###
###################################################################################################

"""
Correlated Pattern Bicluster (CPB) algorithm. Finds biclusters with genes
that have large pairwise Pearson correlation.
"""
from __future__ import division

import os
import subprocess
import tempfile, shutil, numpy
from bicluster import Bicluster, filter_biclusters

BINARY = 'cpb'

def cpb(data, nclus, target_pcc, row_to_col_ratio=1, min_seed_rows=2, max_seed_rows=None, fixed_row=None, fixed_col=None, use_grid_init = True):
    """
    @param float[][] data: 2 dimensinal numpy.array format to represent the data matrix.
    @param int nclus: number of clusters to return. (each is randomly seeded)
    @param float target_pcc: PCC threshold for each row
    @param int min_seed_rows: min rows in the seed. Default is 2.
    @param int max_seed_rows: max rows in the seed.
    @param int fixed_row: The row that is required to be in the seed. -1 for random choice.
    @param int fixed_col: The column that is required to be in the seed. -1 for random choice.
    """
    kwargs = locals()
    if max_seed_rows is None:
        kwargs['max_seed_rows'] = data.shape[0]
    if kwargs['fixed_row'] is None:
        kwargs['fixed_row'] = -1
    if kwargs['fixed_col'] is None:
        kwargs['fixed_col'] = -1
    biclusters = wrapper_caller(BINARY, _write_dataset, _read_results, _do_call, **kwargs)
    return biclusters 

def wrapper_caller(binary, write_dataset, read_results, do_call, data, *args, **kwargs):
    """
    Perform the biclustering.
        
    Takes care of creating a temporary directory, exporting the dataset, running, and reading
    back the results.
    Convenience function for wrappers. 
    
    Many biclustering implementations require specific data formats and produce results in specialized
    formats. Call function for such algorithms, implementing write_dataset(), read_results(), and
    do_call().

    Note: These are the type signatures required of the function arguments.
    def do_call(data, datafile, results_dir, **kwargs)

    def write_dataset(data, filename)

    def read_results(results_dir, data)

    """
    if _which(binary) is None:
        raise WrapperException("executable '{0}' not found on $PATH".format(binary))

    #get a temporary directory to hold dataset and results
    directory = tempfile.mkdtemp()

    #write the dataset in the appropriate format
    datafile = os.path.join(directory, "data.txt")
    write_dataset(data, datafile)

    #prepare location to hold results
    results_dir = os.path.join(directory, "results")
    os.mkdir(results_dir)

    try:
        do_call(data, datafile, results_dir, *args, **kwargs)
    except subprocess.CalledProcessError:
        print ("Return code was not 0. Assuming no biclusters found.")
        sys.exit(1)

    #read back the results
    biclusters = read_results(results_dir, data)

    #cleanup the temporary directory
    shutil.rmtree(directory)
    return biclusters

def _do_call(data, datafile, results_dir, **kwargs):
    datafile = os.path.abspath(datafile)
    results_dir = os.path.abspath(results_dir)

    #write initial bicluster file
    directory = os.path.split(results_dir)[0]
    initfile = os.path.join(directory, 'initfile.txt')
    nrows, ncols = data.shape
    if not kwargs.pop("use_grid_init"):
        init_command = """init_bicluster {0} {1} {nclus} {min_seed_rows} {max_seed_rows} {fixed_row} {fixed_col} {2}""".format(nrows, ncols, initfile, **kwargs)
        try:
            subprocess.check_call(init_command.split())
        except OSError:
            raise Exception("Error calling 'init_bicluster'. Is it on the PATH?")
    else :  
        _cpb_init(nrows, ncols, kwargs["nclus"], initfile, kwargs["fixed_row"], kwargs["fixed_col"])



    #change to the results directory
    saved_path = os.getcwd()
    try:
        os.chdir(results_dir)
        run_command = """{binary} {0} {1} {row_to_col_ratio} {target_pcc}""".format(datafile, initfile, binary=BINARY, **kwargs)
        subprocess.check_call(run_command.split())

    except OSError:
        raise Exception("Error calling 'cpb'. Is it on the PATH?")

    finally:
        os.chdir(saved_path)


def _write_dataset(data, filename):
    """Writes a dataset to the format that CPB reads:

    [num rows] [num columns]
    [first row's values]
    [second row's values]
    ...
    [last row's values]

    """
    outfile = file(filename, 'w')
    rows, cols = data.shape
    outfile.write("{0} {1}".format(rows, cols))
    for row in data:
        outfile.write("\n")
        outfile.write(" ".join([str(j) for j in row]))
    outfile.write("\n")
    outfile.close()


def _read_results(results_dir, data):
    #find the results file with extension .out
    files = os.listdir(results_dir)
    outfiles = [f for f in files if os.path.splitext(f)[1] == '.out']

    biclusters = []
    for f in outfiles:
        biclusters.append(_read_result_file(os.path.join(results_dir, f), data))

    return biclusters



def _read_result_file(filename, data):
    """
    Reads the bicluster in a single CPB output file.

    The file format is:

    ROWS
    [row index]     [row score]
    [row index]     [row score]
    ...
    [row index]     [row score]
    COLS
    [col index]     [col score]
    [col index]     [col score]
    ...
    [col index]     [col score]

    """
    rows, cols = [], []
    with open (filename, 'r') as f:
        target = rows
        for line in f:
            if line[0] == 'R':
                continue
            elif line[0] == 'C':
                target = cols
                continue
            else:
                target.append(int(line.split()[0]))
        rows.sort()
        cols.sort()
    return Bicluster(rows, cols, data=data)



def _which(program):
    """
    Check for an executable on the PATH.

    Taken from http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
 
    """
    import os
    def is_exe(fpath):
        return os.path.exists(fpath) and os.access(fpath, os.X_OK)
 
    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
 
    return None

def _write_list(liste, f):
    f.write(str(liste[0]))
    for i in range(1,len(liste)):
        f.write(" ")
        f.write(str(liste[i]))
    f.write("\n")

def _cpb_init(num_rows, num_cols, num_bic, output_dir, ref_row=-1, ref_col=-1, col_partition = 2, row_partition = 10):

    grid_rows = int(num_rows / row_partition)
    grid_cols = int(num_cols / col_partition)

    rows = range(num_rows)
    cols = range(num_cols)
    f = open(output_dir, "w")
    bic_no_in_pass = (col_partition * row_partition)
    for i in range( int(num_bic / bic_no_in_pass)):
        numpy.random.shuffle(rows)
        numpy.random.shuffle(cols)
        for j in range(row_partition):
            row_start = j * grid_rows
            if j == row_partition - 1:
                row_end = -1
            else:
                row_end = (j+1) * grid_rows
            for k in range (col_partition):
                col_start = k * grid_cols
                if j == col_partition - 1:
                    col_end = -1
                else:
                    col_end = (k+1) * grid_cols

                f.write("BICLUSTER " + str(i * bic_no_in_pass + j * col_partition + k) + "\n")
                f.write(str(ref_row) + "\n")
                _write_list(rows[row_start:row_end],f)
                f.write(str(ref_col) + "\n")
                _write_list(cols[col_start:col_end],f)

    f.close()

