/*
---------------------------------------------------------------------
 This file is part of CPB (Correlated Pattern Biclustering Algorithm)
 Copyright (c) 2011, 
 By:    Doruk Bozdag, 
        Umit V. Catalyurek,
        Mehmet Deveci,
        Onur Kucuktunc,
        Kemal Eren
---------------------------------------------------------------------
 For license info, please see the README.txt and LICENSE.txt files in the main directory.
---------------------------------------------------------------------
*/

#ifndef _cpb_h
#define _cpb_h
#include "cpb.h"
#endif

int read_init_bicluster (FILE *fp, Row_Struct *row, Column_Struct *col, int nrow, int ncol,
                         int *introw, int *nintrow, int *intcol, int *nintcol,
                         int *nfixrow, int *nfixcol);


/*################################################################################*/
/* MAIN ##########################################################################*/
int main (int argc, char *argv[])
{
    FILE *fp, *ocfp; /* Output cluster file handle */
    char *fname, *basename;
    int i, j, k, m, cid;    
    /* INPUT PARAMETERS */
    double r2cR; /* multiplier for row-to-column ratio in the cluster */
    double target_r;
    double fixw = 0;
    int min_ncol = 10;
    /* EXPRESSION MATRIX AND FIXED INFO */
    double **M;
    int nrow, ncol;
    int nfixrow, nfixcol; /* number of fixed rows and columns */
    /* INITIAL CLUSTERS */
    int *introw, *intcol, nintrow, nintcol;
    /* OTHER STRUCTURES */
    Column_Struct *col;
    Row_Struct *row;
    Sort_Struct *col_errs, *row_errs;
    /* TIMING PROFILES */
    clock_t times[10][2];

    times[0][0] = clock();

    /*############################################################*/
    /* READ INPUT PARAMETERS #####################################*/   
    if (argc < 5) {
        printf("Usage: %s <input matrix file> <initial biclusters file> <row-to-column ratio> <target PCC> [%% weight of fixed rows]\n", argv[0]);    
        exit(1);
    }
    r2cR = atof(argv[3]);
    target_r = atof(argv[4]);                
    if (argc >= 6) 
        fixw = atof(argv[5]);
            
    /*############################################################*/
    /* READ EXPRESSION MATRIX AND FIX INFO, AND ALLOCATE ARRAYS ##*/
    times[1][0] = clock();
    /* Read expression matrix */
    read_dbl_matrix(argv[1], &nrow, &ncol, &M, 1); 

#ifdef _DEBUG_INP
    printf("Read matrix -- nrow: %d, ncol: %d\n", nrow, ncol);
#endif
    
    /* Global sized, one time allocations */
    introw = (int *) malloc(sizeof(int) * nrow);
    intcol = (int *) malloc(sizeof(int) * ncol);
    row = (Row_Struct *) malloc(sizeof(Row_Struct) * nrow);
    col = (Column_Struct *) malloc(sizeof(Column_Struct) * ncol);
    fname = (char *) malloc(sizeof(char) * 100);
    basename = (char *) malloc(sizeof(char) * 100);
    col_errs = (Sort_Struct *) malloc(sizeof(Sort_Struct) * ncol);
    row_errs = (Sort_Struct *) malloc(sizeof(Sort_Struct) * nrow);

    /* Get basename of input file */
    for (i=0, j=0; argv[1][i] != '\0'; i++) { 
        if (argv[1][i] == '/') 
            j=0;
        else
            basename[j++] = argv[1][i];
    }
    basename[j] = '\0';    
    times[1][1] = clock();
#ifdef _DEBUG_TIME
    printf ("%8.1lf: Read input hypergraph\n",
            (times[1][1]-times[1][0])*1000.0/CLOCKS_PER_SEC);
#endif
    

    /*############################################################*/
    /* REFINE EACH INITIAL CLUSTER ###############################*/   
    times[4][0] = clock();
    fp = fopen(argv[2], "r");
    for (cid=0; ; cid++) {

#ifdef _DEBUG
        printf("\n\nCLUSTER ID: %d\n", cid);
#endif
        
        /*############################################################*/
        /* CALCULATE AND IMPROVE FITNESS OF EACH CHILD CLUSTER #######*/   
        times[8][0] = clock();

        if (!read_init_bicluster(fp,row,col,nrow,ncol,introw,&nintrow,intcol,&nintcol,&nfixrow,&nfixcol))
            break;
            
        Improve_Fitness(cid, introw, &nintrow, intcol, &nintcol, row, col, M, nrow, ncol,
                        row_errs, col_errs, min_ncol, r2cR, target_r, nfixrow, nfixcol, fixw);
        times[8][1] = clock();

        /*############################################################*/
        /* WRITE CLUSTERS WITH PCC GREATER THAN target_pcc INTO FILES */   
        times[9][0] = clock();
        sprintf(fname, "%s.pcc%g.ran%d.out", basename, target_r, cid);
        ocfp = fopen(fname, "w");
        fprintf(ocfp, "ROWS\n");
        for (k=0; k<nintrow; k++) {
            i = introw[k];
            row_errs[k].idx = i;
            row_errs[k].val = fabs(row[i].r);
        }
        Sort_wrt_vals(row_errs, nintrow);            
        for (k=nintrow-1; k>=0; k--) {            
            i = row_errs[k].idx;
            if (fabs(row[i].r) >= target_r || row[i].fixed)
                fprintf (ocfp, "%-5d %5.3lf\n", i, row[i].r);
        }
        fprintf(ocfp, "COLS\n");
        for (m=0; m<nintcol; m++) {
            j = intcol[m];
            col_errs[m].idx = j;
            col_errs[m].val = col[j].x;
        }
        Sort_wrt_vals(col_errs, nintcol);            
        for (m=0; m<nintcol; m++) {
            j = col_errs[m].idx;
            fprintf (ocfp, "%-5d %7.5lf\n", j, col[j].x);
        }
        fclose(ocfp);

#if 1
        /* Print the returned bicluster as a submatrix */
        sprintf(fname, "%s.pcc%g.ran%d.smx", basename, target_r, cid);
        ocfp = fopen(fname, "w");
        for (k=nintrow-1; k>=0; k--) {            
            i = row_errs[k].idx;
            if (fabs(row[i].r) >= target_r || row[i].fixed) {
                for (m=0; m<nintcol; m++) {
                    j = col_errs[m].idx;
                    fprintf(ocfp, "%g ", M[i][j]);
                }
                fprintf(ocfp, "\n");
            }
        }
        for (m=0; m<nintcol; m++) { // Append cluster center as the last row of the submatrix
            j = col_errs[m].idx;
            fprintf (ocfp, "%g ", col[j].x);
        }
        fprintf(ocfp, "\n");        
        fclose(ocfp);
#endif
        
        times[9][1] = clock();


        
        
#ifdef _DEBUG_TIME
        printf ("        %8.1lf: Improve fitness\n        %8.1f: Write cluster\n",
                (times[8][1]-times[8][0])*1000.0/CLOCKS_PER_SEC,
                (times[9][1]-times[9][0])*1000.0/CLOCKS_PER_SEC);
#endif
        
    }
    fclose(fp);
    times[4][1] = clock();
    
    
#ifdef _DEBUG_TIME
    printf ("    %8.1lf: Refinement of all parts\n", (times[4][1] - times[4][0])*1000.0/CLOCKS_PER_SEC);
#endif

    free(row);
    free(col);
    free(introw);
    free(intcol);
    for (i=0; i<nrow; i++) 
        free(M[i]);
    free(M);
    free(row_errs);
    free(col_errs);
    free(fname);
    free(basename);
    times[0][1] = clock();

    printf ("%8.1lf: TOTAL TIME\n", (times[0][1] - times[0][0])*1000.0/CLOCKS_PER_SEC);
#ifdef _DEBUG_TIME
    printf ("%8.1lf: TOTAL TIME\n", (times[0][1] - times[0][0])*1000.0/CLOCKS_PER_SEC);
#endif
    
    return 0;
}






/*#####################################################################*/
/* READ FIXED ROW AND COLUMN INFO #####################################*/
int read_init_bicluster (FILE *fp, Row_Struct *row, Column_Struct *col, int nrow, int ncol,
                         int *introw, int *nintrow, int *intcol, int *nintcol, int *nfixrow, int *nfixcol)
{
    char *line;
    int *intline, nelt;
    int i, j;

    line = (char *) malloc(MAXLINE * sizeof(char));
    intline = (int *) malloc(MAXLINE/2 * sizeof(int));

    *nfixrow = *nfixcol = *nintcol = *nintrow = 0;
    for (i=0; i<nrow; i++) 
        row[i].fixed = 0;
    for (j=0; j<ncol; j++)
        col[j].fixed = 0;   
                   
    for (i=0; i<5; i++) {
        if (!(fgets(line, MAXLINE-1, fp))) {
            free(line);
            free(intline);
            return 0;
        }
        if (i == 0) /* Skip cluster info line*/
            continue; 
        nelt = str2ia(line, intline, MAXLINE/2);        
        if (i == 1 && intline[0] != -1) {
            *nfixrow = 1;
            row[intline[0]].fixed = 1;
            introw[(*nintrow)++] = intline[0];
        } else if (i == 2) {
            for (j=0; j<nelt; j++)
                introw[(*nintrow)++] = intline[j];        
        } else if (i == 3 && intline[0] != -1) {
            *nfixcol = 1;
            col[intline[0]].fixed = 1;
            intcol[(*nintcol)++] = intline[0];
        } else if (i == 4) { // don't use just else
            for (j=0; j<nelt; j++)
                intcol[(*nintcol)++] = intline[j];
        }
    }    
    free(line);
    free(intline);

#ifdef _DEBUG_INP
    printf("Initial cluster:\n  Rows: ");
    for (j=0; j<*nintrow; j++) 
        printf("%d%s", introw[j], row[introw[j]].fixed ? "* " : " ");
    printf("\n  Cols: ");
    for (j=0; j<*nintcol; j++) 
        printf("%d%s", intcol[j], col[intcol[j]].fixed ? "* " : " ");
    printf("\n");
#endif
    
    return 1;            
}


    


