/*
---------------------------------------------------------------------
 This file is part of CPB (Correlated Pattern Biclustering Algorithm)
 Copyright (c) 2011, 
 By:    Doruk Bozdag, 
        Umit V. Catalyurek,
        Mehmet Deveci,
        Onur Kucuktunc,
        Kemal Eren
---------------------------------------------------------------------
 For license info, please see the README.txt and LICENSE.txt files in the main directory.
---------------------------------------------------------------------
*/

#ifndef _cpb_h
#define _cpb_h
#include "cpb.h"
#endif

/*#########################################################################*/
/* */
void Improve_Fitness(int cid, int *introw, int *rnintrow, int *intcol, int *rnintcol, 
                     Row_Struct *row, Column_Struct *col, double **M, int nrow, int ncol,
                     Sort_Struct *row_errs, Sort_Struct *col_errs, int min_ncol,
                     double r2cR, double target_r, int nfixrow, int nfixcol, double fixw)
{
    int b, c;
    int i, j, k, m, f;
    clock_t times[10][2];
    double max_err, r_min;
    int max_ncol, min_nrow = 2;
    int nstep = 5, max_a = 25, max_b = 20, old_nintcol, old_nintrow, ncol_lim;
    double pcc_stepsize, max_slope_change = 0.05, rm_ratio = 0.5, ncol_stepsize;
    int nintrow = *rnintrow, nintcol = *rnintcol;
    int turn = 0; /* Start with row movement */    
    
    times[0][0] = clock();            
    /*########################################################################*/
    /* INITIALIZATIONS */    
    /* Initialize beta and alpha for each internal row */
    for (k=0; k<nintrow; k++) { 
        i = introw[k]; 
        row[i].beta = 1; 
        row[i].alpha = 0;
    }

    r_min = 9./10*target_r; //2./3*target_r; /* dbdb temporary */
    pcc_stepsize = (target_r - r_min) / (nstep-1);
    ncol_stepsize = (nintcol - min_ncol) / ((double)nstep-1);
    
    for (c=0, ncol_lim=nintcol; c<nstep; c++, r_min += pcc_stepsize, ncol_lim = (int)(min_ncol+(nstep-1-c)*ncol_stepsize)) {
#ifdef _DEBUG
        printf("\n\nThresholds updated, step no: %d/%d, PCC_thr: %.3lf, ncol_thr: %d\n", c, nstep-1, r_min, ncol_lim);
#endif
        
        for (b=0, old_nintcol = 0, old_nintrow = 0; b<max_b; b++) { 

#ifdef _DEBUG
            printf("\n\nEntering b_loop iteration: %d/%d\n", b, max_b-1);
#endif
            
            /*###############################################################################*/
            /* ITERATE TO COMPUTE X FOR INTERNAL COLUMNS AND LEAST SQUARES FOR INTERNAL ROWS */
            Converge_Cluster (nintcol, intcol, &nintrow, introw, row, col, M, max_a, max_slope_change, row_errs, rm_ratio, nfixrow, fixw);
                        
            /*###############################################################################*/
            /* COMPUTE LEAST SQUARES FOR ALL ROWS, X FOR ALL COLS AND ERRORS FOR EITHER #####*/
            if (turn == 0) {
                for (i=0; i<nrow; i++) 
                    Least_Squares_PCC (i, M, nintcol, intcol, row, col);
            } else {                
                for (j=0; j<ncol; j++) {
                    col[j].x = 0;
                    for (k=0; k<nintrow; k++) {
                        i = introw[k];
                        col[j].x += (M[i][j] - row[i].alpha) / row[i].beta;
                    }
                    col[j].x /= nintrow;
                }
            }            
            Compute_Error (M, nrow, ncol, introw, nintrow, intcol, nintcol, row, col, row_errs, col_errs, turn, nfixrow, fixw);
            
#ifdef _DEBUG2
            if (turn == 0) {
                print_row_info(introw, nintrow, row, "After convergence, row PCCs updated\n");
                print_row_errors(nrow, row_errs, "Errors for rows\n", row);
            } else {
                print_col_info(intcol, nintcol, col, "After convergence, column x-values updated\n");
                print_col_errors(ncol, col_errs, "Errors for columns\n");
            }
#endif

            /*###############################################################################*/
            /* MOVE ROWS AND COLUMNS IN/OUT OF THE CLUSTER WRT PCC AND ERRORS ###############*/
            if (turn == 0) {
                /* Make (only) rows with pcc greater than r_min internal */
                for (k=nintrow=f=0; k<nrow; k++) {            
                    i = row_errs[k].idx;
                    if (fabs(row[i].r) < r_min && nintrow >= min_nrow) /* Make sure there are at least min_nrow rows */
                        break;
                    introw[nintrow++] = i;
                    max_err = row_errs[k].val;
                    if (row[i].fixed) /* Count included fixed rows */
                        f++;
                }
                /* Include any remaining fixed row */
                for ( ; f<nfixrow; k++) { 
                    i = row_errs[k].idx;
                    if (row[i].fixed) {
                        introw[nintrow++] = i;
                        f++;
                    }
                }
            } else {            
                /* Determine cols to be made internal. This depends on r2cR and error of row with smallest pcc */
                for (m=max_ncol=0; m<ncol; m++) {
                    if (col_errs[m].val > max_err && max_ncol >= ncol_lim) /* Make sure there are at least ncol_lim cols */
                        break;
                    max_ncol = (int)(r2cR*(m+1));
                }
                max_ncol = max_ncol > ncol ? ncol : max_ncol;
                /* Make (only) selected columns internal */
                for (m=nintcol=f=0; m<max_ncol; m++) {
                    j = col_errs[m].idx;
                    intcol[nintcol++] = j;
                    if (col[j].fixed) /* Count included fixed cols */
                        f++;
                }
                /* Include any remaining fixed column */
                for ( ; f<nfixcol; m++) { 
                    j = col_errs[m].idx;
                    if (col[j].fixed) {
                        intcol[nintcol++] = j;
                        f++;
                    }
                }
            }
            
#ifdef _DEBUG2
            if (turn == 0) 
                print_row_info(introw, nintrow, row, "After row movements\n");
            else 
                print_col_info(intcol, nintcol, col, "After column movements\n");
#endif                

            /* End loop if cluster size changed less than max_size_change */
            if (nintcol == old_nintcol && nintrow == old_nintrow)
                break;
            old_nintcol = nintcol;
            old_nintrow = nintrow;
            turn = turn == 0 ? 1 : 0; /* Toggle turn */
        }
        
#ifdef _DEBUG
        printf("EXITING b loop\n");
#endif
        
    }
    Converge_Cluster (nintcol, intcol, &nintrow, introw, row, col, M, max_a, max_slope_change, row_errs, rm_ratio, nfixrow, fixw); /* Just to make output look more accurate */
    times[0][1] = clock();            

    
    
#if 0       
    printf("clusterid: %d, time: %8.1lf", cid, (times[0][1] - times[0][0])*1000.0/CLOCKS_PER_SEC);    
#ifdef _DEBUG_TIME
    printf ("                 clusterid: %d, fitness: %.2lf\n", clusterid, fitness);
    printf ("            %8.1lf: Internal vtx gain computation\n\
            %8.1lf: External vtx gain computation\n             \
            %8.1lf: Priority queue initialization\n\
            %8.1lf: Vertex movement\n",
           (times[0][1] - times[0][0])*1000.0/CLOCKS_PER_SEC,
           (times[1][1] - times[1][0])*1000.0/CLOCKS_PER_SEC,
           (times[2][1] - times[2][0])*1000.0/CLOCKS_PER_SEC,
           (times[3][1] - times[3][0])*1000.0/CLOCKS_PER_SEC);
#endif

#endif //0

    *rnintrow = nintrow;
    *rnintcol = nintcol;
}





/*################################################################################*/
/* ITERATE TO CONVERGE X FOR INTERNAL COLUMNS AND LEAST SQUARES FOR INTERNAL ROWS */
void Converge_Cluster (int nintcol, int *intcol, int *rnintrow, int *introw,
                       Row_Struct *row, Column_Struct *col, double **M,
                       int max_a, double max_slope_change, Sort_Struct *row_errs, double rm_ratio, int nfixrow, double fixw)
{
    int a, m, j, k, i, end_loop;
    double minx, maxx, rangex, slope_change;
    double xfix;
    int nintrow = *rnintrow;

    if (fixw == 0) /* A fixed row has the same weight as any other internal row */
        fixw = (double)nfixrow/(double)nintrow;
    
    do {
        /* Number of iterations should be larger for larger initial clusters */            
        for(a=0, end_loop = 0; a<max_a && !end_loop; a++) {        
            /* Compute x for each internal column */
            minx = 1.0;
            maxx = 0.0;
            for (m=0; m<nintcol; m++) {
                j = intcol[m];
                col[j].x = 0;
                xfix = 0;
                for (k=0; k<nintrow; k++) {
                    i = introw[k];
                    if (row[i].fixed) 
                        xfix += (M[i][j] - row[i].alpha) / row[i].beta;
                    else
                        col[j].x += (M[i][j] - row[i].alpha) / row[i].beta;
                }
                if (nintrow == nfixrow) 
                    col[j].x = xfix;
                else {                
                    col[j].x /= (nintrow-nfixrow);                    
                    xfix = nfixrow>0 ? xfix/nfixrow : 0;
                    col[j].x = col[j].x*(1-fixw) + xfix*fixw;
                }
                if (m == 0)
                    minx = maxx = col[j].x;
                else if (col[j].x > maxx) 
                    maxx = col[j].x;
                else if (col[j].x < minx)
                    minx = col[j].x;
            }            
            /* Scale x range for each internal column */
            rangex = maxx - minx;
            for (m=0; m<nintcol; m++) {
                j = intcol[m];
                col[j].x = (col[j].x - minx) / rangex;
            }                    
            /* Recompute least squares line for each internal row */
            for (k=0, end_loop = 1; k<nintrow; k++) {                    
                slope_change = Least_Squares_PCC (introw[k], M, nintcol, intcol, row, col);
                if (end_loop && slope_change > max_slope_change) 
                    end_loop = 0;
            }
#ifdef _DEBUG3
            print_row_info(introw, nintrow, row, "Inside Slope Iterations\n");        
            print_col_info(intcol, nintcol, col, "");
#endif                            
        }


        if (!end_loop) { /* Move the non fixed row with smallest pcc out if not converged */

#ifdef _DEBUG2
            printf("Converge_Cluster function iterated %d times. LOOP LIMIT EXCEEDED.\n", a);
#endif
            
            for (k=0, j=0; k<nintrow; k++) { /* Sort internal rows with respect to pcc */
                i = introw[k];
                if (!row[i].fixed) {
                    row_errs[j].idx = i;
                    row_errs[j].val = fabs(row[i].r);
                    j++;
                }
            }
            Sort_wrt_vals(row_errs, j);
            if (j == 0) /* If only fixed rows remain in the cluster, can't do anything more */
                break;
            else {
                int nremove = (int)(j * rm_ratio); /* Number or rows to remove */
                if (nremove > 0) { /* Remove rm_ratio portion of the non-fixed internal rows */
                    for (k=0; k<nremove; k++) { 
                        i = row_errs[j-1-k].idx;
                        for (m=0; m<nintrow; m++) {
                            if (introw[m] == i) {
                                introw[m] = introw[--nintrow]; /* Replace with last element */
                                break;
                            }
                        }
                    }
                } else { /* Remove just 1 row in this case */
                    i = row_errs[j-1].idx;
                    for (m=0; m<nintrow; m++) {
                        if (introw[m] == i) {
                            introw[m] = introw[--nintrow]; /* Replace with last element */
                            break;
                        }
                    }
                }
            }
        }
    } while(end_loop == 0);

    *rnintrow = nintrow;
    
#ifdef _DEBUG2
    printf("Converge_Cluster function iterated %d times. CONVERGED.\n", a);
#endif

}

/*###########################################################################*/
/* COMPUTE LEAST SQUARES LINE AND PCC   #####################################*/
double Least_Squares_PCC (int i, double **M, int nintcol, int *intcol, Row_Struct *row, Column_Struct *col)
{
    int j, m;
    double sum_x, sum_xx, sum_y, sum_yy, sum_xy, oldbeta;
    
    sum_x = sum_xx = sum_y = sum_yy = sum_xy = 0;
    for (m=0; m<nintcol; m++) {
        j = intcol[m];
        sum_x  += col[j].x;
        sum_xx += col[j].x * col[j].x;
        sum_y  += M[i][j]; 
        sum_yy += M[i][j] * M[i][j]; 
        sum_xy += col[j].x * M[i][j];
    }
    oldbeta = row[i].beta;            
    row[i].beta = (sum_xy - sum_x*sum_y/nintcol) / (sum_xx - sum_x*sum_x/nintcol);
    row[i].alpha = (sum_y - row[i].beta*sum_x) / nintcol;
    row[i].r = (sum_xy - sum_x*sum_y/nintcol) / sqrt((sum_xx - sum_x*sum_x/nintcol)*(sum_yy - sum_y*sum_y/nintcol));    
    if (sum_xx == sum_x*sum_x/nintcol || sum_yy == sum_y*sum_y/nintcol) {
        row[i].r = 0;
        row[i].beta = 1;
    }
    return fabs((row[i].beta - oldbeta) / row[i].beta);
}


/*###########################################################################*/
/* COMPUTE AND SORT NORMALIZED RMS ERROR FOR ALL ROWS AND COLUMNS ###########*/
int Compare_vals(const void *A, const void *B)
{
    Sort_Struct *a=(Sort_Struct *)A;
    Sort_Struct *b=(Sort_Struct *)B;    
    if(a->val > b->val) return 1;
    if(a->val < b->val) return -1;
    return 0;
}
void Sort_wrt_vals(Sort_Struct *errs, int nelts)
{
    qsort(errs, nelts, sizeof(Sort_Struct), Compare_vals);
}
void Compute_Error (double **M, int nrow, int ncol, int *introw, int nintrow, int *intcol, int nintcol,
                    Row_Struct *row, Column_Struct *col, Sort_Struct *row_errs, Sort_Struct *col_errs,
                    int turn, int nfixrow, double fixw)
{
    double err, toterr, fixerr;
    int i, j, k, m;
    
    if (fixw == 0) /* A fixed row has the same weight as any other internal row */
        fixw = (double)nfixrow/(double)nintrow;

    if (turn == 0) { /* Compute error for rows only */
        for (i=0; i<nrow; i++) {
            toterr = 0;
            for (m=0; m<nintcol; m++) {
                j = intcol[m];
                err =  col[j].x - (M[i][j] - row[i].alpha) / row[i].beta;
                toterr += err * err;
            }
            toterr = sqrt(toterr) / nintcol;
            row_errs[i].idx = i;
            row_errs[i].val = toterr;
        }
        Sort_wrt_vals(row_errs, nrow);
    } else { /* Compute error for columns only */
        for (j=0; j<ncol; j++) {
            toterr = 0;
            fixerr = 0;
            for (k=0; k<nintrow; k++) {
                i = introw[k];
                err =  col[j].x - (M[i][j] - row[i].alpha) / row[i].beta;
                if (row[i].fixed) 
                    fixerr += err * err;
                else 
                    toterr += err * err;
            }
            if (nintrow == nfixrow)
                toterr = sqrt(fixerr) / nfixrow;
            else if (nfixrow == 0) 
                toterr = sqrt(toterr) / nintrow;
            else {
                toterr = sqrt(toterr)/(nintrow-nfixrow);
                fixerr = sqrt(fixerr)/nfixrow;
                toterr = toterr*(1-fixw) + fixerr*fixw;
            } 
            col_errs[j].idx = j;
            col_errs[j].val = toterr;
        }
        Sort_wrt_vals(col_errs, ncol);
    }
}




#ifdef _DEBUG2
/*###########################################################################*/
/* FUNCTIONS TO PRINT STRUCTURE FIELDS ######################################*/
void print_row_info(int *introw, int nintrow, Row_Struct *row, char *comment)
{
    int i, k;
    printf("%s", comment);
    for (k=0; k<nintrow; k++) {
        i = introw[k];
        printf("row %d: pcc: %.2lf slope: %.2lf, range: %g - %g\n",
               i, row[i].r, row[i].beta, row[i].alpha, row[i].alpha+row[i].beta);
    }
}

void print_col_info(int *intcol, int nintcol, Column_Struct *col, char *comment)
{
    int j, m;
    printf("%s", comment);
    for (m=0; m<nintcol; m++) {
        j = intcol[m];
        printf("col %d: x: %.2lf\n", j, col[j].x);
    }
}

void print_row_errors(int nelts, Sort_Struct *errs, char *comment, Row_Struct *row)
{
    int i;
    printf("%s", comment);
    for (i=0; i<(nelts<10?nelts:20); i++) 
        printf("[%3d]: %.3lf  pcc: %.2lf\n", errs[i].idx, errs[i].val, row[errs[i].idx].r);    
}

void print_col_errors(int nelts, Sort_Struct *errs, char *comment)
{
    int i;
    printf("%s", comment);
    for (i=0; i<(nelts<20?nelts:20); i++) 
        printf("[%3d]: %.3lf\n", errs[i].idx, errs[i].val);    
}

#endif













        

