###################################################################################################
###---------------------------------------------------------------------------------------------###
### This file is part of CPB (Correlated Pattern Biclustering Algorithm)                        ###
### Copyright (c) 2011,                                                                         ###
### By:    Mehmet Deveci,                                                                       ###
###        Onur Kucuktunc,                                                                      ###
###        Kemal Eren,                                                                          ###
###        Doruk Bozdag,                                                                        ###
###        Umit V. Catalyurek                                                                   ###
###                                                                                             ###
###---------------------------------------------------------------------------------------------###
### For license info, please see the README.txt and LICENSE.txt files in the main directory.    ###
###---------------------------------------------------------------------------------------------###
###################################################################################################

"""
Script to run CPB on any given data matrix with the parameters provided.
"""
import sys
from cpb import cpb
from bicluster import filter_biclusters, write_biclusters
import numpy
import shuffle as sh

bic_file = "bic_out"
matrix_file = "matrix_out"
alg_fun = "alg_fun"
start_row = "dsr"
start_col = "dsc"
max_over = "mo"

def print_usage():
    """
    Gives the descripton for the usage of the run_cpy.py script
    """
    print "<usage>: python run_cpb.py <argument_list>"
    print "Example: python run_algorithm.py BF=bic.txt DF=data.txt NB=2"
    print "##########################################################################################################"
    print "argument_list is the list of arguments which are given in the form of <ArgumentSpecifier>=<ArgumentValue>"
    print "##########################################################################################################"
    print "List of arguments:"
    print "##########################################################################################################"
    print "BF : File path for the list of output biclusters that will be returned by the algorithm. - Mandatory Argument"
    print "DF : File path for the input datafile. - Mandatory Argument"
    print "\tDSR : Data start row index - Default is 0 (first row)"
    print "\tDSC : Data start column index - Default is 0 (first column)"
    print "CPB- Correlated Pattern Biclustering"
    print "\tNB : Number of biclusters that the algorithm is expected to find - Mandatory Argument (Should be high enough to cover dataset.)"
    print "\tPCC: PCC threshold for each row - Mandatory Argument"
    print "\tMIR: Min rows in the seed - Default is 2."
    print "\tMAR: Max rows in the seed."
    print "\tRR : Required row in the seed. -1 for random choice(reference row) "
    print "\tRC : Required columns in the seed. -1 for random choice"
    print "\tMO : Max overlap ratio. Between 0-1, Default is 0.75"
    print "\tUGI : To use grid initialization - 0 (Default), or totally random initialization 1."


def parse_argument(arg):
    """
    Parses the given argument string and returns the correct parameter name with value. 
    @param string arg: parameter string given to to script from the command line. It is a string, which is in the <ArgumentSpecifier>=<ArgumentValue> format in valid case.
    """
    try :
        argument = arg.split("=")
        if(not len(argument) == 2 or argument[0] == '' or argument[1] == ''):
            raise
        param_name = argument[0]
        param_value = argument[1]
        if param_name == "BF":
            return bic_file,param_value
        elif param_name == "DF":
            return matrix_file,param_value
        elif param_name == "DSR":
            val = int(param_value)
            if val < -1:
                raise
            return start_row,val
        elif param_name == "DSC":
            val = int(param_value)
            if val < -1:
                raise
            return start_col,val
        elif param_name == "NB":
            val = int(param_value)
            if val < 1:
                raise    
            return "nclus",val          
        elif param_name == "PCC": 
            val = float(param_value)             
            if val < 0 or val > 1:
                raise            
            else:
                return "target_pcc", val
        elif param_name == "MIR":
            val = int(param_value)
            if val < 1:
                raise
            return "min_seed_rows",val
        elif param_name == "MAR":
            val = int(param_value)
            if val < 1:
                raise
            return "max_seed_rows",val
        elif param_name == "RR":
            val = int(param_value)
            if val < -1:
                raise
            return "fixed_row",val
        elif param_name == "RC":
            val = int(param_value)
            if val < -1:
                raise
            return "fixed_col",val
        elif param_name == "MO":
            val = float(param_value)
            if val < 0 or val > 1:
                raise
            return max_over, val
        elif param_name == "UGI":
            val = int(param_value)
            if val == 1: 
                val = True
            elif val == 0:
                val = False
            else:
                raise
            return "use_grid_init", val
        else:
            raise        
    except Exception:
        print "Invalid argument: " + arg
        raise

def get_parameters(arg_list):
    """
    Parses the run command and returns the dictionary of the arguments provided.
    In case of error, exits, and prints the correct usage.
    @param string[] arg_list: list of parameter strings given to to script from the command line. It is a list of strings, which are in the <ArgumentSpecifier>=<ArgumentValue> format in valid case.
    """
    try:
        if(len(arg_list) == 0):
            print_usage()
            raise
        
        dataset_args = dict()
        for arg in arg_list:
            parameter, value = parse_argument(arg)
            dataset_args[parameter] = value
        return dataset_args
    except Exception:
        sys.exit()
    
    
def filter_parameters(parameters):
    """
    Processes the parameters.
    @param dictionary parameters: dictionary of parameters with their values assigned to them. Dictionary in the format <ParameterSpecifier>=<ParameterValue>    
    """
    
    try :
        if not(parameters.has_key(bic_file) and  parameters.has_key(matrix_file)):
            print "BF and DF are mandatory parameters."
            raise
        if not(parameters.has_key("nclus") and  parameters.has_key("target_pcc")):
            print "NB and PCC are mandatory parameters."
            raise

        inputFormat = dict()     
        if parameters.has_key(max_over):
            mo = parameters.pop(max_over)
        else:
            mo = 0.75    
        if not inputFormat.has_key(start_row):
            inputFormat[start_row] = 0
        if not inputFormat.has_key(start_col):
            inputFormat[start_col] = 0            
        bf = parameters.pop(bic_file)
        mf = parameters.pop(matrix_file)
        return parameters, bf,mf, inputFormat, mo
    except Exception:
        sys.exit()
        

def is_number(s):
    """
    Checks if the given s string is an integer or not
    @param string s: string value which will be checked whether a number or not.
    """
    try:
        float(s)
        return True
    except ValueError:
        return False

def readData(dataFile, startCol, startRow):
    """
    Reads data from given dataFile, which is the path. 
    
    @param string dataFile: specifies the path to the data matrix.
    @param int startCol: specifies the start column index for data matrix.
    @param int startRow: specifies the start row index for data matrix.
    """
    try:
        dFile = open(dataFile,'r')
    except IOError:
        print "Could not open data file!"
        sys.exit()
    try:
        for i in range (startRow):        
            dFile.readline()
        data = []
        for line in dFile:
            line = line.split(" ")
            expLine = []
            for exprVal in line[startCol:]:
                if is_number(exprVal):
                    expLine.append(float(exprVal))
                else:
                    expLine.append(0)
            data.append(expLine)
        dFile.close()
        data = numpy.array(data)
        return data
    except IOError:
        print "Error reading the data file"
        sys.exit()

call_cpb = lambda data: cpb(data, **params)

#parameters assignments
parameters = get_parameters(sys.argv[1:])
params, bic_file, matrix_file, inputFormat, max_overlap = filter_parameters(parameters)

#reading the data matrix
data = readData(matrix_file, inputFormat[start_row],  inputFormat[start_col])

#running cpb on shuffled matrix
min_bic = sh.random_max(data,call_cpb)

#running cpb on actual matrix
found = call_cpb(data)

#filtering the results
found = filter_biclusters(found, max_overlap=max_overlap, min_rows=len(min_bic.rows))

#write to output file
write_biclusters(found, bic_file)






