/*
---------------------------------------------------------------------
 This file is part of CPB (Correlated Pattern Biclustering Algorithm)
 Copyright (c) 2011, 
 By:    Doruk Bozdag, 
        Umit V. Catalyurek,
        Mehmet Deveci,
        Onur Kucuktunc,
        Kemal Eren
---------------------------------------------------------------------
 For license info, please see the README.txt and LICENSE.txt files in the main directory.
---------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include "priorityq.h"


/**********************************************************************************************/
/* Create a binary heap */
void Heap_Create(heap *a, int maxsize) { 
    a->size = 0; 
    a->maxsize = maxsize;
    a->nodes = (node *) malloc(sizeof(node) * maxsize);
}

/**********************************************************************************************/
/* Insert an element into the heap */
void Heap_Insert(heap *a, int id, double key, int *hmap)
{
    int i;
    node newnode;

    newnode.id = id;
    newnode.key = key;

    if (a->size >= a->maxsize) {
        printf("Fatal error, heap_insert(): Insufficient heap size. Please allocate larger heap.\n");
        exit(1);
    }

    i = (a->size)++;

    /* Find the a proper place for the newnode and insert there */
    while (i > 0 && newnode.key > (a->nodes[PARENT(i)]).key) {
        a->nodes[i] = a->nodes[PARENT(i)];
        hmap[a->nodes[i].id] = i;
        i = PARENT(i);        
    }
    a->nodes[i] = newnode;
    hmap[a->nodes[i].id] = i;
}

/**********************************************************************************************/
/* Takes a heap rooted at the given index and make sure that is confroms to the heap critera. */
void Heapify(heap *a, int i, int *hmap)
{
    int l, r, largest;
    
    l = LEFT(i);
    r = RIGHT(i);
    
    /* check the left child */
    largest = (l < a->size && (a->nodes[l]).key > (a->nodes[i]).key) ? l : i;
    /* check the right child */
    largest = (r < a->size && (a->nodes[r]).key > (a->nodes[largest]).key) ? r : largest;

    if (largest != i) { 
        /* swap nodes largest and i, then heapify */
        SWAP(a->nodes[i], a->nodes[largest]);
        hmap[a->nodes[i].id] = i;
        hmap[a->nodes[largest].id] = largest;
        Heapify(a, largest, hmap);
    }
}

/**********************************************************************************************/
/* Delete a node from the heap. */
void Heap_Delete(heap *a, int i, int *hmap)
{
    node oldnode;
  
    if (i >= a->size || i < 0) {    
        printf("Fatal error, heap_delete(): Attempted to delete an invalid node. i:%d, heap_size: %d", i, a->size);
        exit(1);
    }
    
    oldnode = a->nodes[i];
    a->nodes[i] = a->nodes[--(a->size)];
    hmap[a->nodes[i].id] = i;        
    Heapify(a, i, hmap);
}

/**********************************************************************************************/
/* Remove the node with largest key from the heap and return it. */
int Heap_Extract_Max(heap *a, node *node0, int *hmap)
{
    if (a->size > 0) {
        *node0 = a->nodes[0];
        a->nodes[0] = a->nodes[--(a->size)];
        hmap[a->nodes[0].id] = 0;
        Heapify(a, 0, hmap);
        return 1;
    } else
        return 0;
}

/**********************************************************************************************/
/* Return the node with largest key without removing it from the heap */
int Heap_Max(heap *a, node *node0)
{
    if (a->size > 0) {
        *node0 = a->nodes[0];
        return 1;
    } else
        return 0;
}

/**********************************************************************************************/
/* Change the key value of a node */
void Heap_Change_Key(heap *a, int i, double newkey, int *hmap)
{
    if (i >= a->size || i < 0) {    
        printf("Fatal error, Heap_Change_Key(): Attempted to edit an invalid node. i:%d, heap_size: %d\n", i, a->size);
        exit(1);
    }
    
    if (newkey < a->nodes[i].key) {
        a->nodes[i].key = newkey;
        Heapify(a, i, hmap);
    } else {
        node tempnode = a->nodes[i];
        tempnode.key = newkey;
        while (i > 0 && newkey > a->nodes[PARENT(i)].key) {
            a->nodes[i] = a->nodes[PARENT(i)];
            hmap[a->nodes[i].id] = i;
            i = PARENT(i);
        }
        a->nodes[i] = tempnode;
        hmap[a->nodes[i].id] = i;
    }    
}

/**********************************************************************************************/
/* Destroy the heap */
void Heap_Destroy (heap *a)
{
    free(a->nodes);
}


